/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// A generic procedure for the computation of chain contraction
/// of a filtered cell complex stored in a file.
/// This is a complete procedure good for simplicial, cubical,
/// or other complexes.
/// It computes an algebraic topological model.
/// This computation fails if homology is not torsion-free.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: August 21, 2013.


#ifndef _ATMODCOMP_H_
#define _ATMODCOMP_H_


// include some standard C++ header files
#include <istream>
#include <ostream>

// include selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/struct/hashsets.h"

// include relevant local header files
#include "chaincon/cellnames.h"
#include "chaincon/filtcomplex.h"
#include "chaincon/readfcompl.h"
#include "chaincon/boundary.h"
#include "chaincon/atmodel.h"
#include "chaincon/homcohom.h"


// --------------------------------------------------
// ---------------- compute AT model ----------------
// --------------------------------------------------

/// Computes an AT model and the homology of a cellular complex.
/// The cellular complex is read from the given file.
/// Only top-dimensional cells should be listed in the file,
/// or otherwise the filtration may not be generated correctly.
template <class CellT, class LinMap, class CellNames>
inline void computeAlgTopModel (const tFilteredComplex<CellT> &K,
	bool relativeComplex,
	chomp::homology::hashedset<CellT> &H,
	LinMap &pi, LinMap &incl, LinMap &phi,
	CellNames &cellNames,
	bool displayPi, bool displayIncl, bool displayPhi,
	bool displayRepr, bool verify, int algorithmVersion)
{
	using chomp::homology::sout;

	// determine the type of coefficients (for displaying homology only)
	typedef typename LinMap::CoefType CoefT;

	// compute the homology together with the chain contraction
	if (relativeComplex)
	{
		algTopModel (K, H, pi, incl, phi, algorithmVersion, K);
	}
	else
	{
		algTopModel (K, H, pi, incl, phi, algorithmVersion,
			Unrestricted ());
	}

	// show the computed homology groups
	chomp::homology::hashedset<CellT> emptyCellArray;
	chomp::homology::hashedset<CoefT> emptyCoefArray;
	std::vector<std::vector<CellT> > representants;
	std::vector<int> bettiNumbers;
	std::vector<std::vector<CoefT> > torsion;
	getHomCohom (H, emptyCellArray, emptyCellArray, emptyCoefArray,
		&representants, &bettiNumbers, &torsion, false);
	sout << "Homology groups:\n";
	sout << homcohom2text (bettiNumbers, torsion, "H_", false);

	// prepare correct cell names and show them if requested to
	if (displayRepr)
		sout << "Representants of homology generators:\n";
	int reprSize = representants. size ();
	for (int dim = 0; dim < reprSize; ++ dim)
	{
		int size = representants [dim]. size ();
		for (int i = 0; i < size; ++ i)
		{
			const CellT &cell (representants [dim] [i]);
			if (displayRepr)
			{
				sout << cellNames (cell) << ": " <<
					cell << ".\n";
			}
			else
			{
				cellNames (cell);
			}
		}
	}

	// show the computed maps
	tCellOwnNames<CellT> cellOwnNames;
	if (displayPi)
	{
		sout << "The projection map pi:\n" <<
			cells2names (pi, cellOwnNames, cellNames);
	}
	if (displayIncl)
	{
		sout << "The inclusion map incl:\n" <<
			cells2names (incl, cellNames, cellOwnNames);
	}
	if (displayPhi)
	{
		sout << "The homology gradient vector field phi:\n" << phi;
	}

	// prepare the boundary map, to be used for verifications
	LinMap boundaryMap;

	// do several verifications to make sure the computed maps are good
	if (verify)
	{
		if (relativeComplex)
			computeBoundaryMap (K, boundaryMap, K);
		else
			computeBoundaryMap (K, boundaryMap, Unrestricted ());
		if (phi * boundaryMap * phi == phi)
			sout << "Verified: 'phi bd phi = phi'.\n";
		else
			sout << "Failed to verify that "
				"'phi pd phi = phi'.\n";

		if (boundaryMap * phi * boundaryMap == boundaryMap)
			sout << "Verified: 'bd phi bd = bd'.\n";
		else
			sout << "Failed to verify that 'bd phi bd = bd'.\n";

		LinMap zeroMap;
		if (phi * phi == zeroMap)
			sout << "Verified: 'phi phi = 0'.\n";
		else
			sout << "Failed to verify that 'phi phi = 0'.\n";

		LinMap composition (boundaryMap * phi);
		composition += phi * boundaryMap;
		composition. negate ();
		addIdentity (K, composition);
		if (incl * pi == composition)
			sout << "Verified: 'i pi = id - bd phi - phi bd'.\n";
		else
			sout << "Failed to verify that "
				"'i pi = id - bd phi - phi bd'.\n";

		LinMap idH;
		addIdentity (H, idH);
		if (pi * incl == idH)
			sout << "Verified: 'pi i = id_H'.\n";
		else
			sout << "Failed to verify that 'pi i = id_H'.\n";

		if (pi * phi == zeroMap)
			sout << "Verified: 'pi phi = 0'.\n";
		else
			sout << "Failed to verify that 'pi phi = 0'.\n";

		if (phi * incl == zeroMap)
			sout << "Verified: 'phi i = 0'.\n";
		else
			sout << "Failed to verify that 'phi i = 0'.\n";
	}

	return;
} /* computeAlgTopModel */


#endif // _ATMODCOMP_H_

