/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// Algebraic topological model computation: An interface to all the
/// algorithms available in this package.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: January 9, 2013.


#ifndef _CHAINCON_ATMODEL_H_
#define _CHAINCON_ATMODEL_H_


// include relevant local header files
#include "chaincon/atmodel0.h"
#include "chaincon/atmodel1.h"
#include "chaincon/atmodel2.h"
#include "chaincon/atmodel2r.h"
#include "chaincon/atmodel3.h"
#include "chaincon/atmodel3r.h"
#include "chaincon/atmodel4r.h"


// --------------------------------------------------
// -------------------- AT model --------------------
// --------------------------------------------------

/// Computes a combinatorial algebraic topological model for a given
/// filtered finite cell complex "K".
/// This function is an interface to the various algorithms
/// available in this package, numbered from 0 to 3.
/// Cells that represent homology generators are appended to the vector "H".
/// The projection map "pi", the inclusion from "H" to the complex "K",
/// and the homology gradient vector field "phi"
/// are assumed to be initially zero and are constructed.
template <class CellT, class CellArray1, class CellArray2, class CellRestrT>
inline void algTopModel (const CellArray1 &K, CellArray2 &H,
	tCombLinMap<CellT,CellT> &pi,
	tCombLinMap<CellT,CellT> &incl,
	tCombLinMap<CellT,CellT> &phi, int version,
	const CellRestrT &restr)
{
	switch (version)
	{
	case 0:
		algTopModel0 (K, H, pi, incl, phi, restr);
		break;
	case 1:
		algTopModel1 (K, H, pi, incl, phi, restr);
		break;
	case 2:
		algTopModel2 (K, H, pi, incl, phi, restr);
		break;
	case 3:
		algTopModel3 (K, H, pi, incl, phi, restr);
		break;
	case 4:
		throw "Algorithm 4 not implemented "
			"for the combinatorial setting.\n"
			"Please, use Algorithm 2 or 3 instead, "
			"or Algorithm 4 with Z_2.";
	default:
		throw "Wrong homology algorithm requested.";
	}
	return;
} /* algTopModel */

/// Computes an algebraict topological model for a given
/// filtered finite cell complex "K" for coeficients in a commutative ring.
/// Cells that represent homology generators are appended to the vector "H".
/// The projection map "pi", the inclusion from "H" to the complex "K",
/// and the homology gradient vector field "phi"
/// are assumed to be initially zero and are constructed.
template <class CellT, class CoefT, class CellArray1, class CellArray2,
	class CellRestrT>
inline void algTopModel (const CellArray1 &K, CellArray2 &H,
	tLinMap<CellT,CellT,CoefT> &pi,
	tLinMap<CellT,CellT,CoefT> &incl,
	tLinMap<CellT,CellT,CoefT> &phi, int version,
	const CellRestrT &restr)
{
	switch (version)
	{
	case 0:
		throw "Algorithm 0 not implemented for Z_p. "
			"Please, use algorithm 2 or 3.";
	case 1:
		throw "Algorithm 1 not implemented for Z_p. "
			"Please, use algorithm 2 or 3.";
	case 2:
		algTopModel2 (K, H, pi, incl, phi, restr);
		break;
	case 3:
		algTopModel3 (K, H, pi, incl, phi, restr);
		break;
	case 4:
		algTopModel4 (K, H, pi, incl, phi, restr);
		break;
	default:
		throw "Wrong homology algorithm requested.";
	}
	return;
} /* algTopModel */


#endif // _CHAINCON_ATMODEL_H_

