/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// A combinatorial tensor (for coefficients in Z_2).
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: January 3, 2013.


#ifndef _CHAINCON_COMBTENSOR_H_
#define _CHAINCON_COMBTENSOR_H_


// include some standard C++ header files
#include <istream>
#include <ostream>

// include selected header files from the CHomP library
#include "chomp/system/config.h"

// include relevant local header files
#include "chaincon/pair.h"
#include "chaincon/combchain.h"
#include "chaincon/ringz2.h"


// --------------------------------------------------
// -------------- combinatorial tensor --------------
// --------------------------------------------------

/// Combinatorial tensor of cells. This is in fact a list of pairs of cells
/// of the given type. One can add products of chains to the tensor,
/// which is then decomposed into tensors of individual cells and reduced.
template <class Cell1T, class Cell2T>
class tCombTensor
{
public:
	/// The type of cells on the left in the combinatorial tensor.
	typedef Cell1T Cell1Type;

	/// The type of cells on the right in the combinatorial tensor.
	typedef Cell2T Cell2Type;

	/// The type of coefficients in the chain.
	typedef tZ2 CoefType;

	/// The default constructor of an empty combinatorial tensor.
	tCombTensor ();

	/// Returns the number of elements in the combinatorial tensor.
	int_t size () const;

	/// Returns true if and only if the tensor is zero (empty set).
	bool empty () const;

	/// Returns the n-th left element of the combinatorial tensor.
	const Cell1T &left (int_t n) const;

	/// Returns the n-th right element of the combinatorial tensor.
	const Cell2T &right (int_t n) const;

	/// Adds a pair of cells to the combinatorial tensor.
	void add (const Cell1T &c1, const Cell2T &c2);

	/// Adds a pair of chains to the combinatorial tensor.
	void add (const tCombChain<Cell1T> &c1,
		const tCombChain<Cell2T> &c2);

	/// Adds a given tensor to the combinatorial tensor.
	tCombTensor<Cell1T,Cell2T> &operator +=
		(const tCombTensor<Cell1T,Cell2T> &ch);

	/// Compares if the two tensors are equal.
	bool operator == (const tCombTensor<Cell1T,Cell2T> &ch) const;

	/// Swaps the data with another tensor.
	void swap (tCombTensor<Cell1T,Cell2T> &ch);

private:
	/// The set of elements in the combinatorial tensor.
	tCombChain<tPair<Cell1T,Cell2T> > tensor;

}; /* class tCombTensor */

// --------------------------------------------------

template <class Cell1T, class Cell2T>
inline tCombTensor<Cell1T,Cell2T>::tCombTensor ()
{
	return;
} /* tCombTensor::tCombTensor */

template <class Cell1T, class Cell2T>
inline int_t tCombTensor<Cell1T,Cell2T>::size () const
{
	return tensor. size ();
} /* tCombTensor::size */

template <class Cell1T, class Cell2T>
inline bool tCombTensor<Cell1T,Cell2T>::empty () const
{
	return tensor. empty ();
} /* tCombTensor::empty */

template <class Cell1T, class Cell2T>
inline const Cell1T &tCombTensor<Cell1T,Cell2T>::left (int_t n) const
{
	return tensor. getCell (n). left;
} /* tCombTensor::left */

template <class Cell1T, class Cell2T>
inline const Cell2T &tCombTensor<Cell1T,Cell2T>::right (int_t n) const
{
	return tensor. getCell (n). right;
} /* tCombTensor::right */

template <class Cell1T, class Cell2T>
inline void tCombTensor<Cell1T,Cell2T>::add
	(const Cell1T &c1, const Cell2T &c2)
{
	tensor. add (tPair<Cell1T,Cell2T> (c1, c2));
} /* tCombTensor::add */

template <class Cell1T, class Cell2T>
inline void tCombTensor<Cell1T,Cell2T>::add
	(const tCombChain<Cell1T> &ch1, const tCombChain<Cell2T> &ch2)
{
	int_t ch1size = ch1. size ();
	int_t ch2size = ch2. size ();
	for (int_t i = 0; i < ch1size; ++ i)
	{
		const Cell1T &c1 = ch1. getCell (i);
		for (int_t j = 0; j < ch2size; ++ j)
		{
			const Cell2T &c2 = ch2. getCell (j);
			tensor. add (tPair<Cell1T,Cell2T> (c1, c2));
		}
	}
	return;
} /* tCombTensor::add */

template <class Cell1T, class Cell2T>
inline tCombTensor<Cell1T,Cell2T> &tCombTensor<Cell1T,Cell2T>::operator +=
	(const tCombTensor<Cell1T,Cell2T> &ch)
{
	tensor += ch. tensor;
	return *this;
} /* tCombTensor::operator += */

template <class Cell1T, class Cell2T>
inline bool tCombTensor<Cell1T,Cell2T>::operator ==
	(const tCombTensor<Cell1T,Cell2T> &ch) const
{
	return (tensor == ch. tensor);
} /* tCombTensor::operator == */

template <class Cell1T, class Cell2T>
inline void tCombTensor<Cell1T,Cell2T>::swap (tCombTensor<Cell1T,Cell2T> &ch)
{
	tensor. swap (ch. tensor);
	return;
} /* tCombTensor::swap */

// --------------------------------------------------

/// Writes a combinatorial tensor to an output stream.
template <class Cell1T, class Cell2T>
std::ostream &operator << (std::ostream &out,
	const tCombTensor<Cell1T,Cell2T> &c)
{
	int_t size = c. size ();
	if (!size)
		out << '0';
	else
	{
		for (int_t i = 0; i < size; ++ i)
		{
			if (i)
				out << " + ";
			out << c. left (i) << " x " << c. right (i);
		}
	}
	return out;
} /* operator << */

/// Reads a combinatorial tensor from an input stream.
template <class Cell1T, class Cell2T>
std::istream &operator >> (std::istream &in, tCombTensor<Cell1T,Cell2T> &c)
{
	throw "Operator >> not implemented for tCombTensor.";
	return in;
} /* operator >> */


#endif // _CHAINCON_COMBTENSOR_H_

