/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// The Shih operator as a component
/// of the Eilenberg-Zilber chain contraction.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: December 4, 2014.


#ifndef _CHAINCON_EZ_SHI_H_
#define _CHAINCON_EZ_SHI_H_


// include some standard C++ header files
#include <istream>
#include <ostream>

// include selected header files from the CHomP library
#include "chomp/system/config.h"
//#include "chomp/system/textfile.h"

// include relevant local header files
#include "chaincon/combtensor.h"
#include "chaincon/tensor.h"
#include "chaincon/prodcell.h"
#include "chaincon/shuffles.h"


// --------------------------------------------------
// ------------------ SHI operator ------------------
// --------------------------------------------------

/// Computes the Shih operator on a product of two simplicial cells.
/// Combinatorial version (coefficients in Z_2).
template <class SimCellT, class ProdCellT>
inline void EZ_SHI (const SimCellT &s1, const SimCellT &s2,
	tCombChain<ProdCellT> &t)
{
	int m (s1. dim ());
	if (m != s2. dim ())
		throw "Different dimensions of cells for the SHI operator.";

	for (int q = 0; q < m; ++ q)
	{
		for (int p = 0; p < m - q; ++ p)
		{
			Shuffles sh (p + 1, q);
			std::vector<int> alpha (p + 1);
			std::vector<int> beta (q);
			int sig (0);
			while (1)
			{
				int mBar (m - p - q);

				// get the next shuffle
				bool moreAvailable =
					sh. get (alpha, beta, sig);

				// compute the appropriate cells
				SimCellT cell1 (s1);
				for (int i = 0; i < q; ++ i)
				{
					SimCellT cell1new (cell1, m - i);
					cell1. swap (cell1new);
				}
				{
					SimCellT cell1new (cell1,
						-1 - (mBar - 1));
					cell1. swap (cell1new);
				}
				for (int i = 0; i < q; ++ i)
				{
					SimCellT cell1new (cell1,
						-1 - (beta [i] + mBar));
					cell1. swap (cell1new);
				}
				SimCellT cell2 (s2);
				for (int i = 0; i < p; ++ i)
				{
					SimCellT cell2new (cell2,
						m - q - i - 1);
					cell2. swap (cell2new);
				}
				for (int i = 0; i <= p; ++ i)
				{
					SimCellT cell2new (cell2,
						-1 - (alpha [i] + mBar));
					cell2. swap (cell2new);
				}

				// add the product to the result
				ProdCellT prod (cell1, cell2);
				if (!prod. degenerate ())
					t. add (prod);

				// finish if no more shuffles are available
				if (!moreAvailable)
					break;
			}
		}
	}

	return;
} /* EZ_SHI */

/// Computes the Shih operator on a product of two simplicial cells.
template <class SimCellT, class ProdCellT, class CoefT>
inline void EZ_SHI (const SimCellT &s1, const SimCellT &s2,
	tChain<ProdCellT,CoefT> &t)
{
//	using chomp::homology::sbug;

	int m (s1. dim ());
	if (m != s2. dim ())
		throw "Different dimensions of cells for the SHI operator.";

	for (int q = 0; q < m; ++ q)
	{
		for (int p = 0; p < m - q; ++ p)
		{
			Shuffles sh (p + 1, q);
			std::vector<int> alpha (p + 1);
			std::vector<int> beta (q);
			int sig (0);
			while (1)
			{
				int mBar (m - p - q);

				// get the next shuffle
				bool moreAvailable =
					sh. get (alpha, beta, sig);

				// compute the appropriate cells
				SimCellT cell1 (s1);
				for (int i = 0; i < q; ++ i)
				{
					SimCellT cell1new (cell1, m - i);
					cell1. swap (cell1new);
				}
				{
					SimCellT cell1new (cell1,
						-1 - (mBar - 1));
					cell1. swap (cell1new);
				}
				for (int i = 0; i < q; ++ i)
				{
					SimCellT cell1new (cell1,
						-1 - (beta [i] + mBar));
					cell1. swap (cell1new);
				}
				SimCellT cell2 (s2);
				for (int i = 0; i < p; ++ i)
				{
					SimCellT cell2new (cell2,
						m - q - i - 1);
					cell2. swap (cell2new);
				}
				for (int i = 0; i <= p; ++ i)
				{
					SimCellT cell2new (cell2,
						-1 - (alpha [i] + mBar));
					cell2. swap (cell2new);
				}

				// determine the correct coefficient
				CoefT coef (1);
				if ((mBar + sig) & 1)
					coef. negate ();

				// add the product to the result
				ProdCellT prod (cell1, cell2);
				if (!prod. degenerate ())
					t. add (prod, coef);
			//	sbug << "DEBUG EZ_SHI: p = " << p <<
			//		", q = " << q << ": " <<
			//		coef << " * " <<
			//		cell1 << " X " << cell2 << "\n";

				// finish if no more shuffles are available
				if (!moreAvailable)
					break;
			}
		}
	}

	return;
} /* EZ_SHI */

// --------------------------------------------------

/// Computes the Shih operator on a product of two simplicial cells.
/// Combinatorial version (coefficients in Z_2).
template <class ProdCellT>
inline void EZ_SHI (const ProdCellT &s, tCombChain<ProdCellT> &t)
{
	EZ_SHI (s. getLeft (), s. getRight (), t);

	return;
} /* EZ_SHI */

/// Computes the Shih operator on a product of two simplicial cells.
template <class ProdCellT, class CoefT>
inline void EZ_SHI (const ProdCellT &s, tChain<ProdCellT,CoefT> &t)
{
	EZ_SHI (s. getLeft (), s. getRight (), t);

	return;
} /* EZ_SHI */

// --------------------------------------------------

/// Computes the Shih operator on a chain of products of simplicial cells.
/// Combinatorial version (coefficients in Z_2).
template <class ProdCellT>
inline void EZ_SHI (const tCombChain<ProdCellT> &s,
	tCombChain<ProdCellT> &t)
{
	for (int_t n = 0; n < s. size (); ++ n)
	{
		tCombChain<ProdCellT> result;
		EZ_SHI (s. getCell (n), result);
		t += result;
	}

	return;
} /* EZ_SHI */

/// Computes the Shih operator on a chain of products of simplicial cells.
template <class ProdCellT, class CoefT>
inline void EZ_SHI (const tChain<ProdCellT,CoefT> &s,
	tChain<ProdCellT,CoefT> &t)
{
	for (int_t n = 0; n < s. size (); ++ n)
	{
		tChain<ProdCellT,CoefT> result;
		EZ_SHI (s. getCell (n), result);
		result *= s. getCoef (n);
		t += result;
	}

	return;
} /* EZ_SHI */


#endif // _CHAINCON_EZ_SHI_H_

