/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// This file includes a specific interface to an SNF computation algorithm.
/// In the default setting it uses the algorithm from the CHomP package,
/// but you are welcome to use another, hopefully more efficient algorithm.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: June 26, 2012.


#ifndef _CHAINCON_SNF_H_
#define _CHAINCON_SNF_H_


// include some standard C++ header files
#include <algorithm>

// include selected header files from the CHomP library
#include "chomp/system/config.h"

// include relevant local header files
#include "chaincon/filtcomplex.h"
#include "chaincon/chain.h"


// --------------------------------------------------
// --------- the choice of an SNF algorithm ---------
// --------------------------------------------------

#include "chaincon/snfchomp.h"


// --------------------------------------------------
// ---------------- SNF verification ----------------
// --------------------------------------------------

/// Verifies if the change of basis matrices match each other.
/// Displays messages and throws an exception in case of a failure.
template <class MatrixT>
void verifyChangeOfBasisSNF (const MatrixT &matrix, int maxDim)
{
	using chomp::homology::sout;
	using chomp::homology::timeused;

	typedef typename MatrixT::ChainType AlgChainType;
//	typedef typename MatrixT::CoefType CoefType;

	timeused verifTime;
	verifTime = 0;
	sout << "Change of basis check: ";

	for (int q = 0; q < maxDim; ++ q)
	{
		int_t nColumns = matrix. getNumCols (q);
		for (int_t n = 0; n < nColumns; ++ n)
		{
			AlgChainType ch (matrix. getNewChain (q, n));
			AlgChainType result;
			int_t size = ch. size ();
			for (int_t i = 0; i < size; ++ i)
			{
				AlgChainType part (matrix. getOldChain
					(q, ch. getCell (i). getId ()));
				part *= ch. getCoef (i);
				result += part;
			}
			if (!(result. size () == 1) ||
				!(result. getCell (0). getId () == n) ||
				!(result. getCoef (0) == 1))
			{
				throw "Wrong change of basis (1).";
			}
		}
		for (int_t n = 0; n < nColumns; ++ n)
		{
			AlgChainType ch (matrix. getOldChain (q, n));
			AlgChainType result;
			int_t size = ch. size ();
			for (int_t i = 0; i < size; ++ i)
			{
				AlgChainType part (matrix. getNewChain
					(q, ch. getCell (i). getId ()));
				part *= ch. getCoef (i);
				result += part;
			}
			if (!(result. size () == 1) ||
				!(result. getCell (0). getId () == n) ||
				!(result. getCoef (0) == 1))
			{
				throw "Wrong change of basis (2).";
			}
		}
	}
	sout << "Completed in " << verifTime << ".\n";

	return;
} /* verifyChangeOfBasisSNF */


#endif // _CHAINCON_SNF_H_

