/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// A tensor of chains with coefficients in an arbitrary commutative ring.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: April 12, 2013.


#ifndef _CHAINCON_TENSOR_H_
#define _CHAINCON_TENSOR_H_


// include some standard C++ header files
#include <istream>
#include <ostream>

// include selected header files from the CHomP library
#include "chomp/system/config.h"

// include relevant local header files
#include "chaincon/pair.h"
#include "chaincon/chain.h"


// --------------------------------------------------
// --------------------- tensor ---------------------
// --------------------------------------------------

/// Tensor of chains. This is in fact a chain made up of pairs of cells
/// of the given type. One can add products of chains to the tensor,
/// which is then decomposed into tensors of individual cells and reduced.
template <class Cell1T, class Cell2T, class CoefT>
class tTensor
{
public:
	/// The type of left cells in the tensor.
	typedef Cell1T Cell1Type;

	/// The type of right cells in the tensor.
	typedef Cell2T Cell2Type;

	/// The type of coefficients in the chain.
	typedef CoefT CoefType;

	/// The default constructor of an empty tensor.
	tTensor ();

	/// Returns the number of elements in the tensor.
	int_t size () const;

	/// Returns true if and only if the tensor is zero (empty set).
	bool empty () const;

	/// Returns the n-th left element of the tensor.
	const Cell1T &left (int_t n) const;

	/// Returns the n-th right element of the tensor.
	const Cell2T &right (int_t n) const;

	/// Returns the n-th coefficient in the tensor.
	const CoefT &coef (int_t n) const;

	/// Adds a pair of cells to the tensor.
	void add (const Cell1T &c1, const Cell2T &c2, const CoefT &coef);

	/// Adds a pair of chains to the tensor.
	void add (const tChain<Cell1T,CoefT> &c1,
		const tChain<Cell2T,CoefT> &c2);

	/// Adds a given tensor to the tensor.
	tTensor<Cell1T,Cell2T,CoefT> &operator +=
		(const tTensor<Cell1T,Cell2T,CoefT> &ch);

	/// Multiples the tensor by a given coefficient.
	tTensor<Cell1T,Cell2T,CoefT> &operator *= (const CoefT &coef);

	/// Compares if the two tensors are equal.
	bool operator == (const tTensor<Cell1T,Cell2T,CoefT> &ch) const;

	/// Swaps the data with another tensor.
	void swap (tTensor<Cell1T,Cell2T,CoefT> &ch);

private:
	/// The set of elements in the tensor.
	tChain<tPair<Cell1T,Cell2T>,CoefT> tensor;

}; /* class tTensor */

// --------------------------------------------------

template <class Cell1T, class Cell2T, class CoefT>
inline tTensor<Cell1T,Cell2T,CoefT>::tTensor ()
{
	return;
} /* tTensor::tTensor */

template <class Cell1T, class Cell2T, class CoefT>
inline int_t tTensor<Cell1T,Cell2T,CoefT>::size () const
{
	return tensor. size ();
} /* tTensor::size */

template <class Cell1T, class Cell2T, class CoefT>
inline bool tTensor<Cell1T,Cell2T,CoefT>::empty () const
{
	return tensor. empty ();
} /* tTensor::empty */

template <class Cell1T, class Cell2T, class CoefT>
inline const Cell1T &tTensor<Cell1T,Cell2T,CoefT>::left (int_t n) const
{
	return tensor. getCell (n). left;
} /* tTensor::left */

template <class Cell1T, class Cell2T, class CoefT>
inline const Cell2T &tTensor<Cell1T,Cell2T,CoefT>::right (int_t n) const
{
	return tensor. getCell (n). right;
} /* tTensor::right */

template <class Cell1T, class Cell2T, class CoefT>
inline const CoefT &tTensor<Cell1T,Cell2T,CoefT>::coef (int_t n) const
{
	return tensor. getCoef (n);
} /* tTensor::coef */

template <class Cell1T, class Cell2T, class CoefT>
inline void tTensor<Cell1T,Cell2T,CoefT>::add
	(const Cell1T &c1, const Cell2T &c2, const CoefT &coef)
{
	tensor. add (tPair<Cell1T,Cell2T> (c1, c2), coef);
} /* tTensor::add */

template <class Cell1T, class Cell2T, class CoefT>
inline void tTensor<Cell1T,Cell2T,CoefT>::add
	(const tChain<Cell1T,CoefT> &ch1, const tChain<Cell2T,CoefT> &ch2)
{
	int_t ch1size = ch1. size ();
	int_t ch2size = ch2. size ();
	for (int_t i = 0; i < ch1size; ++ i)
	{
		const Cell1T &cell1 = ch1. getCell (i);
		const CoefT &coef1 = ch1. getCoef (i);
		for (int_t j = 0; j < ch2size; ++ j)
		{
			const Cell2T &cell2 = ch2. getCell (j);
			CoefT coef (ch2. getCoef (j));
			coef *= coef1;
			add (cell1, cell2, coef);
		}
	}
	return;
} /* tTensor::add */

template <class Cell1T, class Cell2T, class CoefT>
inline tTensor<Cell1T,Cell2T,CoefT> &
	tTensor<Cell1T,Cell2T,CoefT>::operator +=
	(const tTensor<Cell1T,Cell2T,CoefT> &ch)
{
	tensor += ch. tensor;
	return *this;
} /* tTensor::operator += */

template <class Cell1T, class Cell2T, class CoefT>
inline tTensor<Cell1T,Cell2T,CoefT> &
	tTensor<Cell1T,Cell2T,CoefT>::operator *=
	(const CoefT &coef)
{
	tensor *= coef;
	return *this;
} /* tTensor::operator *= */

template <class Cell1T, class Cell2T, class CoefT>
inline bool tTensor<Cell1T,Cell2T,CoefT>::operator ==
	(const tTensor<Cell1T,Cell2T,CoefT> &ch) const
{
	return (tensor == ch. tensor);
} /* tTensor::operator == */

template <class Cell1T, class Cell2T, class CoefT>
inline void tTensor<Cell1T,Cell2T,CoefT>::swap
	(tTensor<Cell1T,Cell2T,CoefT> &ch)
{
	tensor. swap (ch. tensor);
	return;
} /* tTensor::swap */

// --------------------------------------------------

/// Writes a tensor to an output stream.
template <class Cell1T, class Cell2T, class CoefT>
std::ostream &operator << (std::ostream &out,
	const tTensor<Cell1T,Cell2T,CoefT> &c)
{
	int_t size = c. size ();
	if (!size)
		out << '0';
	else
	{
		for (int_t i = 0; i < size; ++ i)
		{
			if (i)
				out << " + ";
			out << c. coef (i) << " * " <<
				c. left (i) << " x " << c. right (i);
		}
	}
	return out;
} /* operator << */

/// Reads a tensor from an input stream.
template <class Cell1T, class Cell2T, class CoefT>
std::istream &operator >> (std::istream &in,
	tTensor<Cell1T,Cell2T,CoefT> &c)
{
	throw "Operator >> not implemented for tTensor.";
	return in;
} /* operator >> */


#endif // _CHAINCON_TENSOR_H_

