/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// A generic procedure for the computation of the cohomology ring structure
/// for an algebraic minimal model.
/// This is a complete procedure good for simplicial, cubical,
/// or other complexes.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: September 17, 2015.


#ifndef _CRINGCOMP_H_
#define _CRINGCOMP_H_


// include some standard C++ header files
#include <istream>
#include <ostream>
#include <vector>
#include <utility>

// include selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/struct/hashsets.h"
#include "chomp/struct/multitab.h"

// include relevant local header files
#include "chaincon/cellnames.h"
#include "chaincon/linmap.h"
#include "chaincon/pair.h"
#include "chaincon/cring.h"


// --------------------------------------------------
// ------------ compute cohomology ring -------------
// --------------------------------------------------

/// Computes the cohomology ring, given a minimal model of a cell complex,
/// and shows the result.
template <class CellT, class CoefT, class CellArray1, class CellArray2,
	class CellArray3, class CoefArray, class CellNames>
inline void computeCohomRing (const CellArray1 &H, const CellArray2 &A,
	const CellArray3 &B, const CoefArray &Q,
	const tLinMap<CellT,CellT,CoefT> &pi,
	const tLinMap<CellT,CellT,CoefT> &incl,
	CellNames &cellNames, bool verify)
{
	using chomp::homology::sout;
	using chomp::homology::sbug;
	using chomp::homology::timeused;
	using chomp::homology::hashedset;
	using chomp::homology::multitable;

//	typedef typename LinMap::CellDomType CellT;
//	typedef tLinMap<CellT,CellT,CoefT> LinMap;

	// prepare
	timeused compTime;
	compTime = 0;

	// determine cohomology generators
//	sout << "Determining cohomology generators "
//		"from the algebraic minimal model...\n";
	std::vector<std::vector<CellT> > cohomRepresentants;
	std::vector<int> cohomBettiNumbers;
	std::vector<std::vector<CoefT> > cohomTorsion;
	getHomCohom (H, A, B, Q, &cohomRepresentants, &cohomBettiNumbers,
		&cohomTorsion, true);

	// prepare correct cell names (in case it was not done before)
//	sout << "Setting up names of the representants...\n";
	int cohomReprSize = cohomRepresentants. size ();
	for (int dim = 0; dim < cohomReprSize; ++ dim)
	{
		int size = cohomRepresentants [dim]. size ();
		for (int i = 0; i < size; ++ i)
			cellNames (cohomRepresentants [dim] [i]);
	}

	// prepare a data structure for the cup products
	typedef tPair<CellT,CellT> CellPair;
	hashedset<CellPair> cellPairs;
	multitable<tChain<CellT,CoefT> > cupProducts;

	// compute all the cup products
	cring (cohomRepresentants, pi, incl, cellPairs, cupProducts);

	// show the computed nonzero cup products
	// in the ascending order by the dimension of the "big" cells
	bool nonzeroFound = false;
	int_t size = cellPairs. size ();
	for (int_t n = 0; n < size; ++ n)
	{
		if (cupProducts [n]. empty ())
			continue;
		if (!nonzeroFound)
		{
			sout << "The computed nonzero cup products are:\n";
			nonzeroFound = true;
		}
		sout << cellNames (cellPairs [n]. left) << " u " <<
			cellNames (cellPairs [n]. right) << " = " <<
			cells2names (cupProducts [n], cellNames) << ".\n";
	}
	if (!nonzeroFound)
		sout << "All the computed cup products are trivial.\n";

	// show information on the total computation time
	sout << "The ring structure was computed in " << compTime << ".\n";

	return;
} /* computeCohomRing */


#endif // _CRINGCOMP_H_

