#!/usr/bin/perl

#############################################################################
#
# This script creates a suspension of a cubical complex.
#
#############################################################################

# Copyright (C) 2015-2016 by Pawel Pilarczyk.
#
# This file is part of my research software package. This is free software:
# you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file "license.txt". If not,
# please, see <http://www.gnu.org/licenses/>.
#
# Started on December 28, 2015. Last revision: January 19, 2016.

use strict;

print "SuspCubCell 0.01 (C) 2015-2016 by Pawel Pilarczyk.\n";
die "Call with the INPUT and OUTPUT file names.\n" if (scalar (@ARGV) < 2);

my $inName = $ARGV [0];
my $outName = $ARGV [1];

my $file;
my $spaceDim;
my @cubCells;
my @lowestCoords;
my @highestCoords;
open ($file, "<$inName") or die "Can't open '${inName}'.\n";
print "Reading cubical cells from '${inName}...\n";
my $count = 0;
while (my $line = <$file>)
{
	# read the cubical cell from the input file
	next if $line =~ m/^\s*;/;
	die " ERROR: Full cubical cells not yet supported.\n" .
		"\tPlease, encode your cubes " .
		"as cartesian products of intervals.\n"
		if $line =~ m/^\s*[\(\d\-,\s\)]+\s*$/;
	next unless $line =~ m/^\s*([\[\]\d,\s\-x]+)/;
	my $cellRead = $1;
	$cellRead =~ m/(.*?)\s*$/;
	$cellRead = $1;
	next if ($cellRead eq '');

	# determine the intervals whose product defines the cell
	my @intervals;
	if ($cellRead =~ m/x/)
	{
	#	print "DEBUG: '${cellRead}'.\n";
		@intervals = split (/\s*x\s*/, $cellRead);
	}
	else
	{
		print "DEBUG: '${cellRead}'.\n";
		die "Full cubical cells not yet supported, sorry!\n";
	}

	# store the cubical cell in the form of the product of intervals
	my $cubCell = join ('x', @intervals);
	push (@cubCells, $cubCell);

	# update the ranges of coordinates
	for (my $i = 0; $i < scalar (@intervals); ++ $i)
	{
		my $interval = $intervals [$i];
		my $lower;
		my $higher;
		if ($interval =~ m/\[([\d\-]+),([\d\-]+)\]/)
		{
			$lower = $1;
			$higher = $2;
		}
		elsif ($interval =~ m/\[([\d\-]+)\]/)
		{
			$lower = $1;
			$higher = $1;
		}
		else
		{
			die "Can't decode interval: ${interval}.\n";
		}
		$lowestCoords [$i] = $lower if
			(!defined ($lowestCoords [$i])) ||
			($lowestCoords [$i] > $lower);
		$highestCoords [$i] = $higher if
			(!defined ($highestCoords [$i])) ||
			($highestCoords [$i] < $higher);
	}
	++ $count;
}
close $file;
print "${count} cubical cells read.\n";
print "The coordinate ranges are: ";
for (my $i = 0; $i < scalar (@lowestCoords); ++ $i)
{
	print 'x' if $i;
	print "[", $lowestCoords [$i], ",", $highestCoords [$i], "]";
}
print ".\n";

open ($file, ">$outName") or die "Can't create '${outName}'.\n";
print "Writing the cubical suspension to '${outName}'...\n";

# write the cubical complex extended by the product with the interval [0,1]
foreach my $cubCell (@cubCells)
{
	print $file "${cubCell}x[0,1]\n";
}
print $file "\n";

# write the top and bottom rectangles
my @counters;
for (my $i = 0; $i < scalar (@lowestCoords); ++ $i)
{
	$counters [$i] = $lowestCoords [$i];
}
while (1)
{
	# output the top and bottom cells corresponding to the counters
	for (my $layer = 0; $layer <= 1; ++ $layer)
	{
		for (my $i = 0; $i < scalar (@counters); ++ $i)
		{
			print $file 'x' if $i;
			print $file "[", $counters [$i], ",",
				($counters [$i] + 1), "]";
		}
		print $file "x[${layer}]\n";
	}

	# take the next set of counters and quit if done
	my $quit = 1;
	for (my $i = 0; $i < scalar (@counters); ++ $i)
	{
		++ $counters [$i];
		if ($counters [$i] <= $highestCoords [$i])
		{
			$quit = 0;
			last;
		}
		$counters [$i] = $lowestCoords [$i];
	}
	last if $quit;
}
print "Done.\n";

