/////////////////////////////////////////////////////////////////////////////
///
/// \file
///
/// A simple program for showing the shuffles.
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2009-2016 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on March 24, 2009. Last revision: August 21, 2013.


// include some standard C++ header files
#include <vector>

// include selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/system/arg.h"

// include relevant local header files
#include "chaincon/shuffles.h"


// --------------------------------------------------
// -------------------- OVERTURE --------------------
// --------------------------------------------------

/// The title of the program and licensing information.
const char *title = "\
Test Shuffles. August 12, 2013. (C) 1997-2014 by Pawel Pilarczyk.\n\
This is free software. No warranty. Consult 'license.txt' for details.";

/// Brief help information on the program's usage.
const char *helpinfo = "\
This program shows all the shuffles of required size.\n\
Call with:\n\
p q - the two positive integer numbers that define the shuffles.\n\
Switches and additional arguments:\n\
--log filename - save the output to a file (without progress indicators),\n\
--quiet - suppress data output to the screen (whcih can be still logged),\n\
--help - display this brief help information only and exit.\n\
For more information please consult the accompanying documentation\n\
or ask the program's author at http://www.PawelPilarczyk.com/.";


// --------------------------------------------------
// ---------------------- main ----------------------
// --------------------------------------------------

/// The main procedure of the program.
/// Returns: 0 = Ok, -1 = Error, 1 = Help displayed, 2 = Wrong arguments.
int main (int argc, char *argv [])
{
	using namespace chomp::homology;

	// turn on a message that will appear if the program does not finish
	program_time = "Aborted after";

	// prepare user-configurable data
	int p = -1;
	int q = -1;

	// analyze the command line
	arguments a;
	arg (a, NULL, p);
	arg (a, NULL, q);
	arghelp (a);

	argstreamprepare (a);
	int argresult = a. analyze (argc, argv);
	argstreamset ();

	// show the program's title
	if (argresult >= 0)
		sout << title << '\n';

	// if something was incorrect, show an additional message and exit
	if (argresult < 0)
	{
		sout << "Call with '--help' for help.\n";
		return 2;
	}

	// if help requested or no filename present, show help information
	if ((argresult > 0) || (p < 0) || (q < 0))
	{
		sout << helpinfo << '\n';
		return 1;
	}

	// try running the main function and catch an error message if thrown
	try
	{
		Shuffles s (p, q);
		std::vector<int> alpha (p);
		std::vector<int> beta (q);
		int sig (0);
		int count = 0;
		while (1)
		{
			bool moreAvailable = s. get (alpha, beta, sig);
			sout << "Shuffle " << count << ": (";
			for (int i = 0; i < p; ++ i)
				sout << (i ? " " : "") << alpha [i];
			sout << "), (";
			for (int i = 0; i < q; ++ i)
				sout << (i ? " " : "") << beta [i];
			sout << "); sig = " << sig << ".\n";
			++ count;
			if (!moreAvailable)
				break;
		}

		// check if the number of shuffles is correct
		int all = 1;
		for (int i = p + 1; i <= p + q; ++ i)
			all *= i;
		for (int i = 2; i <= q; ++ i)
			all /= i;
		if (count != all)
		{
			sout << "Wrong number of shuffles: " << count <<
				" instead of " << all << ".\n";
		}

		program_time = "Total time used:";
		program_time = 1;
		return 0;
	}
	catch (const char *msg)
	{
		sout << "ERROR: " << msg << '\n';
		return -1;
	}
	catch (const std::exception &e)
	{
		sout << "ERROR: " << e. what () << '\n';
		return -1;
	}
	catch (...)
	{
		sout << "ABORT: An unknown error occurred.\n";
		return -1;
	}
} /* main */

